#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <mdf/mdf.h>
#include <mdf/mdf_fields.h>
#include "common.h"

#if defined _WIN32 || defined _WIN64
# include <winsock2.h>
# include <windows.h>
#else
# include <unistd.h>
#endif

/* this callback function will be called during mdf_connect() so that we can display
 * some progress, could be usefull in a GUI application for example. We also have the
 * possibility of getting notified when heartbeats are sent and/or received so that we
 * can e.g show the network RTT latency on an idle connection. */
static void status_callback (void *userdata, MDF_CONN_STATUS status, const char *host, const char *ip)
{
	static uint64_t last_request = 0;
	time_t now = time (NULL);
	struct tm *tm = gmtime (&now);

	if (status == MDF_STATUS_LOOKUP) {
		printf ("%02d:%02d:%02d [MDF_STATUS_LOOKUP] DNS lookup for host %s\n", tm->tm_hour, tm->tm_min, tm->tm_sec, host);
	} else if (status == MDF_STATUS_CONNECTING) {
		printf ("%02d:%02d:%02d [MDF_STATUS_CONNECTING] Connecting to %s\n", tm->tm_hour, tm->tm_min, tm->tm_sec, ip);
	} else if (status == MDF_STATUS_CONNECTED) {
		printf ("%02d:%02d:%02d [MDF_STATUS_CONNECTED] Connected to %s\n", tm->tm_hour, tm->tm_min, tm->tm_sec, ip);
	} else if (status == MDF_STATUS_DISCONNECTED) {
		printf ("%02d:%02d:%02d [MDF_STATUS_DISCONNECTED] Disconnected from %s\n", tm->tm_hour, tm->tm_min, tm->tm_sec, ip);
	} else if (status == MDF_STATUS_READYTOLOGON) {
		printf ("%02d:%02d:%02d [MDF_STATUS_READYTOLOGON] Handshake with server is complete, you can now logon\n", tm->tm_hour, tm->tm_min, tm->tm_sec);
	} else if (status == MDF_STATUS_SND_HB_REQ) {
		printf ("%02d:%02d:%02d [MDF_STATUS_SND_HB_REQ] Client sent a Heartbeat request to the server: curtime=%s\n", tm->tm_hour, tm->tm_min, tm->tm_sec, host);
		last_request = strtoull (host, NULL, 10);
	} else if (status == MDF_STATUS_RCV_HB_REQ) {
		printf ("%02d:%02d:%02d [MDF_STATUS_RCV_HB_REQ] Client received a Heartbeat request from the server: curtime=%s servertime=%s\n", tm->tm_hour, tm->tm_min, tm->tm_sec, host, ip);
	} else if (status == MDF_STATUS_RCV_HB_RES) {
		/* let's calculate the round-trip network latency from when we sent the request
		 * to we received the reply. This will erroneously display an initial RTT of 0.00ns 
		 * since the server will send an unsolicited reply upon connect. */
		double diff = last_request == 0 ? 0.0 : strtoull (host, NULL, 10) - last_request;

		char suffix[4] = "ns";

		if (diff > 1000.0) {
			diff /= 1000.0;
			strcpy (suffix, "µs");

			if (diff > 1000.0) {
				diff /= 1000.0;
				strcpy (suffix, "ms");

				if (diff > 1000.0) {
					diff /= 1000.0;
					strcpy (suffix, "s");

					if (diff > 60.0) { /* a bit silly I know */
						diff /= 60.0;
						strcpy (suffix, "min");
					}
				}
			}
		}

		printf ("%02d:%02d:%02d [MDF_STATUS_RCV_HB_RES] Client received a Hearbeat response from the server: curtime=%s servertime=%s latency=%.2f%s\n", tm->tm_hour, tm->tm_min, tm->tm_sec, host, ip, diff, suffix);
	}
}

/* send a request to the server for Basic Data and Quotes for the supplied
 * list of instruments. If the list is NULL then we issue a wildcard request
 * i.e make the server send data for all available instruments
 *
 * Please note that we use a condensed request, i.e we issue the basic data
 * and the quote request in a single request. When doing that we have to use
 * the mdf_message_add_list() function and we also have to put a " " between
 * the request classes since the string has to be space separated */
static void send_request (mdf_t mdf, mdf_message_t message, char *request, time_t last_update)
{
	mdf_message_add (message, 0, MDF_M_REQUEST);
	mdf_message_add_list (message, MDF_F_REQUESTCLASS, MDF_RC_BASICDATA " " MDF_RC_QUOTE);
	mdf_message_add_numeric (message, MDF_F_REQUESTTYPE, MDF_RT_FULL);

	if (request != NULL)
		mdf_message_add_list (message, MDF_F_INSREFLIST, request);
	else
		mdf_message_add_string (message, MDF_F_INSREFLIST, "*");

	if (last_update != 0) {
		struct tm *tm;
		int timediff;
		char reqdate[11], reqtime[9];

		/* compensate for clock drift and/or network latency */
		if (!mdf_get_property (mdf, MDF_OPT_TIME_DIFFERENCE, &timediff))
			timediff = 0;

		last_update += timediff;

		/* we decrease by one second to be on the safe side */
		last_update--;
		tm = gmtime (&last_update);

		sprintf (reqdate, "%d-%02d-%02d", tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday);
		sprintf (reqtime, "%02d:%02d:%02d", tm->tm_hour, tm->tm_min, tm->tm_sec);

		mdf_message_add_date (message, MDF_F_DATE, reqdate);
		mdf_message_add_time (message, MDF_F_TIME, reqtime);
		printf ("Requesting data updated since %sT%sZ\n", reqdate, reqtime);
	}

	/* we supply a dummy value "0" for request id just so that we can force
	 * the server to send us a Request Finished message when we have received all
	 * the image (snapshot) data since we use that as the trigger to when we should
	 * start collecting the time of last update. */
	mdf_message_add_numeric (message, MDF_F_REQUESTID, "0");

	mdf_message_send (mdf, message);
	mdf_message_reset (message);
}

int main (int argc, char **argv)
{
	char *server = NULL;
	char *username = NULL;
	char *password = NULL;
	static char server_default[] = "sandbox.millistream.com:9100";
	mdf_t mdf = mdf_create ();
	mdf_message_t message = mdf_message_create ();
	int i;
	int mdf_fd;
	int request_finished;
	time_t last_update = 0;

	for (i = 1; i < argc; i++) {
		if (strcmp (argv[i], "--server") == 0) {
			server = argv[i + 1];

		} else if (strcmp (argv[i], "--username") == 0) {
			username = argv[i + 1];

		} else if (strcmp (argv[i], "--password") == 0) {
			password = argv[i + 1];
		} else {
			printf ("unknown argument: %s\n", argv[i]);
			continue;
		}

		i++;
	}

	if (server == NULL)
		server = server_default;

	if (username == NULL || password == NULL) {
		printf ("usage: example --username <username> --password <password> [--server <server>]\n");
		return 1;
	}

	/* setup the status callback so that we can receive notifications on how mdf_connect() progresses */
	mdf_set_property (mdf, MDF_OPT_STATUS_CALLBACK_FUNCTION, (void *)status_callback);

perform_reconnect:
	if (!mdf_connect (mdf, server)) {
		printf ("unable to connect with server: %s\n", server); 

#if defined _WIN32 || defined _WIN64
		Sleep (10000);
#else
		sleep (10);
#endif
		mdf_disconnect (mdf);
		goto perform_reconnect;
	}

	if (send_logon (mdf, message, "libmdf C advanced example", username, password) != 0) {
		printf ("logon failed\n");

#if defined _WIN32 || defined _WIN64
		Sleep (10000);
#else
		sleep (10);
#endif
		mdf_disconnect (mdf);
		goto perform_reconnect;
	}

	/* we fetch the mdf filedescriptor so we can use select() */
	mdf_get_property (mdf, MDF_OPT_FD, &mdf_fd);

	/* here we send a request for three insrefs:
	 * 772 - 'Ericsson B' which is a Share instrument
	 * 26021 - 'Bats Europe CXE' which is a Market Place instrument
	 * 907169 - 'XCSE Other Equities' which is a Tick Table instrument
	 *
	 * Depending upon your account you might not have access to these
	 * particular instruments! */
	send_request (mdf, message, "772 26021 907169", last_update);

	request_finished = 0;

	for (;;) {
		fd_set rdevents;
		struct timeval tv;
		int ret;

		FD_ZERO (&rdevents);
		FD_SET (mdf_fd, &rdevents);

		/* we ask libmdf how many seconds we can wait in maximum before we
		 * have to call mdf_consume() to have it manage it's internal idle
		 * functionality (e.g sending heartbeats) */
		int timeout;
		mdf_get_property (mdf, MDF_OPT_TIMEOUT, &timeout);

		tv.tv_sec = timeout;
		tv.tv_usec = 0;

		ret = select (mdf_fd + 1, &rdevents, NULL, NULL, &tv);

		/* TODO: we should probably do something else than simply quit here... */
		if (ret == -1) {
			printf ("select() returned -1 on us, we quit!\n");
			break;
		}

		/* we call mdf_consume() if select() has marked a read event for us, but also on
		 * timeout so that the API has a chance of sending heartbeats on idle.
		 *
		 * If we have other descriptors waiting on the same select then we cannot rely on
		 * timeout from select() so in that case we should either check time spent since
		 * we last called mdf_consume() against MDF_OPT_TIMEOUT or we should add say a
		 * timer_fd primed to MDF_OPT_TIMEOUT and call mdf_consume() when that timer 
		 * signals. */
		if (ret == 0 || FD_ISSET (mdf_fd, &rdevents)) {
			/* there is data to consume, since we use select() to tell us this we call
			 * mdf_consume() with "0" as timeout so we don't trigger libmdf to call
			 * select() internally and thus putting us to sleep again, in this example
			 * there would be no consequences other than wasting a syscall but if this
			 * where a real world application where we also listened to data from other
			 * descriptors then this would stall the entire application until there
			 * either where data receoved on the libmdf connection or mdf_consume()
			 * timed out */
			int ret = mdf_consume (mdf, 0);

			if (ret == -1) { /* disconnected, let's wait 10s before we do a reconnect so we don't spam the server */
				printf ("disconnected from feed\n");
#if defined _WIN32 || defined _WIN64
				Sleep (10000);
#else
				sleep (10);
#endif
				mdf_disconnect (mdf);
				goto perform_reconnect;
			} else if (ret == 1) { /* we have data! */
				uint16_t mref;
				uint64_t insref;

				if (request_finished == 1)
					last_update = time (NULL);

				while (mdf_get_next_message2 (mdf, &mref, &insref)) {
					uint32_t tag;
					char *value;

					/* we use this as a signal for when to start collecting the time of the last update,
					 * we do this in order to only begin requests from a point in time when we have started to
					 * receive realtime data for all our requested instruments. */
					if (mref == MDF_M_REQUESTFINISHED)
						request_finished = 1;

					print_header (mref, mdf_get_mclass(mdf), insref);

					while (mdf_get_next_field (mdf, &tag, &value)) {
						print_field (tag, value);
					}

					printf ("\n");
				}
			}
		}
	}

	send_logoff (mdf, message);
	mdf_message_destroy (message);
	mdf_destroy (mdf);
}

