#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <mdf/mdf.h>
#include <mdf/mdf_fields.h>
#include "common.h"

/* this callback function will be called every time there is
 * data to consume */
static void callback (void *userdata, mdf_t handle)
{
	uint16_t mref;
	uint64_t insref;

	while (mdf_get_next_message2 (handle, &mref, &insref)) {
		uint32_t tag;
		char *value;

		print_header (mref, mdf_get_mclass(handle), insref);

		while (mdf_get_next_field (handle, &tag, &value)) {
			print_field (tag, value);
		}

		printf ("\n");
	}
}

/* send a request to the server for Basic Data and Quotes for the supplied
 * list of instruments. If the list is NULL then we issue a wildcard request
 * i.e make the server send data for all available instruments 
 *
 * Please note that we use a condensed request, i.e we issue the basic data
 * and the quote request in a single request. When doing that we have to use
 * the mdf_message_add_list() function and we also have to put a " " between
 * the request classes since the string has to be space separated */
static void send_request (mdf_t mdf, mdf_message_t message, char *request)
{
	mdf_message_add (message, 0, MDF_M_REQUEST);
	mdf_message_add_list (message, MDF_F_REQUESTCLASS, MDF_RC_BASICDATA " " MDF_RC_QUOTE);
	mdf_message_add_numeric (message, MDF_F_REQUESTTYPE, MDF_RT_FULL);

	if (request != NULL)
		mdf_message_add_list (message, MDF_F_INSREFLIST, request);
	else
		mdf_message_add_string (message, MDF_F_INSREFLIST, "*");

	mdf_message_send (mdf, message);
	mdf_message_reset (message);
}

int main (int argc, char **argv)
{
	char *server = NULL;
	char *username = NULL;
	char *password = NULL;
	static char server_default[] = "sandbox.millistream.com:9100";
	mdf_t mdf = mdf_create ();
	mdf_message_t message = mdf_message_create ();
	int i;

	for (i = 1; i < argc; i++) {
		if (strcmp (argv[i], "--server") == 0) {
			server = argv[i + 1];

		} else if (strcmp (argv[i], "--username") == 0) {
			username = argv[i + 1];

		} else if (strcmp (argv[i], "--password") == 0) {
			password = argv[i + 1];
		} else {
			printf ("unknown argument: %s\n", argv[i]);
			continue;
		}

		i++;
	}

	if (server == NULL)
		server = server_default;

	if (username == NULL || password == NULL) {
		printf ("usage: callback_example --username <username> --password <password> [--server <server>]\n");
		return 1;
	}

	if (!mdf_connect (mdf, server)) {
		printf ("unable to connect with server: %s\n", server); 
		return -1;
	}

	if (send_logon (mdf, message, "libmdf C callback example", username, password) != 0) {
		mdf_message_destroy (message);
		mdf_destroy (mdf);
		return 1;
	}

	/* we set the callback function after the call to send_logon(), otherwise
	 * the callback would catch the logon response from the server and we don't
	 * want that (because the send_logon() function uses mdf_consume() directly
	 * and needs to see the Logon reply from the server, if we had registered the
	 * callback before this then the Logon reply would have been sent to our callback
	 * and not to the send_logon() function).
	 *
	 * So if you expand upon this example and keep using the callback functionality
	 * then you have to set it to NULL with mdf_set_property() before you call
	 * send_logon() again after say a disconnect, or change the code so that 
	 * you handle the logon reply in the callback instead. */
	mdf_set_property (mdf, MDF_OPT_DATA_CALLBACK_FUNCTION, (void *)callback);

	/* here we send a request for three insrefs:
	 * 772 - 'Ericsson B' which is a Share instrument
	 * 26021 - 'Bats Europe CXE' which is a Market Place instrument
	 * 907169 - 'XCSE Other Equities' which is a Tick Table instrument
	 *
	 * Depending upon your account you might not have access to these
	 * particular instruments! */
	send_request (mdf, message, "772 26021 907169");

	for (;;) {
		/* we loop until we get disconnected or have a network error */
		if (mdf_consume (mdf, 10) == -1)
			break;
	}

	send_logoff (mdf, message);
	mdf_message_destroy (message);
	mdf_destroy (mdf);
}

